@@ -2,7 +2,7 @@
 
   gdbm.c -
 
-  $Author: yugui $
+  $Author: nobu $
   modified at: Mon Jan 24 15:59:52 JST 1994
 
   Documentation by Peter Adolphs < futzilogik at users dot sourceforge dot net >
@@ -44,7 +44,7 @@
  * 1. Opening/creating a database, and filling it with some entries:
  *
  *      require 'gdbm'
- *      
+ *
  *      gdbm = GDBM.new("fruitstore.db")
  *      gdbm["ananas"]    = "3"
  *      gdbm["banana"]    = "8"
@@ -54,7 +54,7 @@
  * 2. Reading out a database:
  *
  *      require 'gdbm'
- *      
+ *
  *      gdbm = GDBM.new("fruitstore.db")
  *      gdbm.each_pair do |key, value|
  *        print "#{key}: #{value}\n"
@@ -73,12 +73,18 @@
  */
 static VALUE rb_cGDBM, rb_eGDBMError, rb_eGDBMFatalError;
 
+#if SIZEOF_LONG > SIZEOF_INT
+#define TOO_LONG(n) ((long)(+(int)(n)) != (long)(n))
+#else
+#define TOO_LONG(n) 0
+#endif
+
 #define RUBY_GDBM_RW_BIT 0x20000000
 
 #define MY_BLOCK_SIZE (2048)
 #define MY_FATAL_FUNC rb_gdbm_fatal
 static void
-rb_gdbm_fatal(char *msg)
+rb_gdbm_fatal(const char *msg)
 {
     rb_raise(rb_eGDBMFatalError, "%s", msg);
 }
@@ -95,13 +101,13 @@ closed_dbm(void)
 }
 
 #define GetDBM(obj, dbmp) do {\
-    Data_Get_Struct(obj, struct dbmdata, dbmp);\
-    if (dbmp == 0) closed_dbm();\
-    if (dbmp->di_dbm == 0) closed_dbm();\
+    Data_Get_Struct((obj), struct dbmdata, (dbmp));\
+    if ((dbmp) == 0) closed_dbm();\
+    if ((dbmp)->di_dbm == 0) closed_dbm();\
 } while (0)
 
 #define GetDBM2(obj, data, dbm) {\
-    GetDBM(obj, data);\
+    GetDBM((obj), (data));\
     (dbm) = dbmp->di_dbm;\
 }
 
@@ -204,31 +210,40 @@ fgdbm_initialize(int argc, VALUE *argv, VALUE obj)
 
     SafeStringValue(file);
 
+#ifdef GDBM_CLOEXEC
+    /* GDBM_CLOEXEC is available since gdbm 1.10. */
+    flags |= GDBM_CLOEXEC;
+#endif
+
     if (flags & RUBY_GDBM_RW_BIT) {
         flags &= ~RUBY_GDBM_RW_BIT;
-        dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+        dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                         flags, mode, MY_FATAL_FUNC);
     }
     else {
         dbm = 0;
         if (mode >= 0)
-            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                             GDBM_WRCREAT|flags, mode, MY_FATAL_FUNC);
         if (!dbm)
-            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                             GDBM_WRITER|flags, 0, MY_FATAL_FUNC);
         if (!dbm)
-            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE, 
+            dbm = gdbm_open(RSTRING_PTR(file), MY_BLOCK_SIZE,
                             GDBM_READER|flags, 0, MY_FATAL_FUNC);
     }
 
+    if (dbm) {
+        rb_fd_fix_cloexec(gdbm_fdesc(dbm));
+    }
+
     if (!dbm) {
         if (mode == -1) return Qnil;
 
         if (gdbm_errno == GDBM_FILE_OPEN_ERROR ||
             gdbm_errno == GDBM_CANT_BE_READER ||
             gdbm_errno == GDBM_CANT_BE_WRITER)
-            rb_sys_fail(RSTRING_PTR(file));
+            rb_sys_fail_str(file);
         else
             rb_raise(rb_eGDBMError, "%s", gdbm_strerror(gdbm_errno));
     }
@@ -290,17 +305,20 @@ rb_gdbm_fetch(GDBM_FILE dbm, datum key)
     str = rb_str_new(val.dptr, val.dsize);
     free(val.dptr);
     OBJ_TAINT(str);
-    return (VALUE)str;
+    return str;
 }
 
 static VALUE
 rb_gdbm_fetch2(GDBM_FILE dbm, VALUE keystr)
 {
     datum key;
+    long len;
 
     StringValue(keystr);
+    len = RSTRING_LEN(keystr);
+    if (TOO_LONG(len)) return Qnil;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
 
     return rb_gdbm_fetch(dbm, key);
 }
@@ -336,9 +354,12 @@ rb_gdbm_nextkey(GDBM_FILE dbm, VALUE keystr)
 {
     datum key, key2;
     VALUE str;
+    long len;
 
+    len = RSTRING_LEN(keystr);
+    if (TOO_LONG(len)) return Qnil;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
     key2 = gdbm_nextkey(dbm, key);
     if (key2.dptr == 0)
         return Qnil;
@@ -416,9 +437,9 @@ fgdbm_key(VALUE obj, VALUE valstr)
 
         valstr2 = rb_gdbm_fetch2(dbm, keystr);
         if (!NIL_P(valstr2) &&
-            RSTRING_LEN(valstr) == RSTRING_LEN(valstr2) &&
+            (int)RSTRING_LEN(valstr) == (int)RSTRING_LEN(valstr2) &&
             memcmp(RSTRING_PTR(valstr), RSTRING_PTR(valstr2),
-                   RSTRING_LEN(valstr)) == 0) {
+                   (int)RSTRING_LEN(valstr)) == 0) {
             return keystr;
         }
     }
@@ -435,9 +456,9 @@ fgdbm_index(VALUE obj, VALUE value)
 
 /*
  * call-seq:
- *      gdbm.select { |value| block } -> array
+ *      gdbm.select { |key, value| block } -> array
  *
- * Returns a new array of all values of the database for which _block_
+ * Returns a new array of all key-value pairs of the database for which _block_
  * evaluates to true.
  */
 static VALUE
@@ -495,11 +516,14 @@ rb_gdbm_delete(VALUE obj, VALUE keystr)
     datum key;
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
+    long len;
 
     rb_gdbm_modify(obj);
     StringValue(keystr);
+    len = RSTRING_LEN(keystr);
+    if (TOO_LONG(len)) return Qnil;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
 
     GetDBM2(obj, dbmp, dbm);
     if (!gdbm_exists(dbm, key)) {
@@ -537,7 +561,7 @@ fgdbm_delete(VALUE obj, VALUE keystr)
  * call-seq:
  *      gdbm.shift -> (key, value) or nil
  *
- * Removes a key-value-pair from this database and returns it as a 
+ * Removes a key-value-pair from this database and returns it as a
  * two-item array [ _key_, _value_ ]. Returns nil if the database is empty.
  */
 static VALUE
@@ -570,7 +594,7 @@ fgdbm_delete_if(VALUE obj)
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
     VALUE keystr, valstr;
-    VALUE ret, ary = rb_ary_new();
+    VALUE ret, ary = rb_ary_tmp_new(0);
     int i, status = 0, n;
 
     rb_gdbm_modify(obj);
@@ -581,8 +605,9 @@ fgdbm_delete_if(VALUE obj)
     for (keystr = rb_gdbm_firstkey(dbm); RTEST(keystr);
          keystr = rb_gdbm_nextkey(dbm, keystr)) {
 
+	OBJ_FREEZE(keystr);
         valstr = rb_gdbm_fetch2(dbm, keystr);
-        ret = rb_protect(rb_yield, rb_assoc_new(keystr, valstr), &status);
+        ret = rb_protect(rb_yield, rb_assoc_new(rb_str_dup(keystr), valstr), &status);
         if (status != 0) break;
         if (RTEST(ret)) rb_ary_push(ary, keystr);
         GetDBM2(obj, dbmp, dbm);
@@ -591,7 +616,8 @@ fgdbm_delete_if(VALUE obj)
     for (i = 0; i < RARRAY_LEN(ary); i++)
         rb_gdbm_delete(obj, RARRAY_PTR(ary)[i]);
     if (status) rb_jump_tag(status);
-    if (n > 0) dbmp->di_size = n - RARRAY_LEN(ary);
+    if (n > 0) dbmp->di_size = n - (int)RARRAY_LEN(ary);
+    rb_ary_clear(ary);
 
     return obj;
 }
@@ -619,7 +645,7 @@ fgdbm_clear(VALUE obj)
             free(key.dptr);
             rb_raise(rb_eGDBMError, "%s", gdbm_strerror(gdbm_errno));
         }
-        free(key.dptr); 
+        free(key.dptr);
     }
 #else
     while (key = gdbm_firstkey(dbm), key.dptr) {
@@ -683,10 +709,10 @@ fgdbm_store(VALUE obj, VALUE keystr, VALUE valstr)
     StringValue(valstr);
 
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = RSTRING_LENINT(keystr);
 
     val.dptr = RSTRING_PTR(valstr);
-    val.dsize = RSTRING_LEN(valstr);
+    val.dsize = RSTRING_LENINT(valstr);
 
     GetDBM2(obj, dbmp, dbm);
     dbmp->di_size = -1;
@@ -939,10 +965,13 @@ fgdbm_has_key(VALUE obj, VALUE keystr)
     datum key;
     struct dbmdata *dbmp;
     GDBM_FILE dbm;
+    long len;
 
     StringValue(keystr);
+    len = RSTRING_LENINT(keystr);
+    if (TOO_LONG(len)) return Qfalse;
     key.dptr = RSTRING_PTR(keystr);
-    key.dsize = RSTRING_LEN(keystr);
+    key.dsize = (int)len;
 
     GetDBM2(obj, dbmp, dbm);
     if (gdbm_exists(dbm, key))
@@ -973,9 +1002,9 @@ fgdbm_has_value(VALUE obj, VALUE valstr)
         valstr2 = rb_gdbm_fetch2(dbm, keystr);
 
         if (!NIL_P(valstr2) &&
-            RSTRING_LEN(valstr) == RSTRING_LEN(valstr2) &&
+            (int)RSTRING_LEN(valstr) == (int)RSTRING_LEN(valstr2) &&
             memcmp(RSTRING_PTR(valstr), RSTRING_PTR(valstr2),
-                   RSTRING_LEN(valstr)) == 0) {
+                   (int)RSTRING_LEN(valstr)) == 0) {
             return Qtrue;
         }
     }
@@ -1023,6 +1052,7 @@ fgdbm_reorganize(VALUE obj)
     rb_gdbm_modify(obj);
     GetDBM2(obj, dbmp, dbm);
     gdbm_reorganize(dbm);
+    rb_fd_fix_cloexec(gdbm_fdesc(dbm));
     return obj;
 }
 
