@@ -2,7 +2,7 @@
 
   signal.c -
 
-  $Author: usa $
+  $Author: ngoto $
   created at: Tue Dec 20 10:13:44 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -15,34 +15,39 @@
 #include "vm_core.h"
 #include <signal.h>
 #include <stdio.h>
+#include <errno.h>
+#include "ruby_atomic.h"
+#include "eval_intern.h"
 
-#ifdef _WIN32
-typedef LONG rb_atomic_t;
-
-# define ATOMIC_TEST(var) InterlockedExchange(&(var), 0)
-# define ATOMIC_SET(var, val) InterlockedExchange(&(var), (val))
-# define ATOMIC_INC(var) InterlockedIncrement(&(var))
-# define ATOMIC_DEC(var) InterlockedDecrement(&(var))
-
-#else
-typedef int rb_atomic_t;
-
-# define ATOMIC_TEST(var) ((var) ? ((var) = 0, 1) : 0)
-# define ATOMIC_SET(var, val) ((var) = (val))
-# define ATOMIC_INC(var) (++(var))
-# define ATOMIC_DEC(var) (--(var))
+#if defined(__native_client__) && defined(NACL_NEWLIB)
+# include "nacl/signal.h"
 #endif
 
-#ifdef __BEOS__
+#ifdef NEED_RUBY_ATOMIC_OPS
+rb_atomic_t
+ruby_atomic_exchange(rb_atomic_t *ptr, rb_atomic_t val)
+{
+    rb_atomic_t old = *ptr;
+    *ptr = val;
+    return old;
+}
+
+rb_atomic_t
+ruby_atomic_compare_and_swap(rb_atomic_t *ptr, rb_atomic_t cmp,
+			     rb_atomic_t newval)
+{
+    rb_atomic_t old = *ptr;
+    if (old == cmp) {
+      *ptr = newval;
+    }
+    return old;
+}
+#endif
+
+#if defined(__BEOS__) || defined(__HAIKU__)
 #undef SIGBUS
 #endif
 
-#if defined HAVE_SIGPROCMASK || defined HAVE_SIGSETMASK
-#define USE_TRAP_MASK 1
-#else
-#define USE_TRAP_MASK 0
-#endif
-
 #ifndef NSIG
 # define NSIG (_SIGMAX + 1)      /* For QNX */
 #endif
@@ -208,6 +213,26 @@ signo2signm(int no)
     return 0;
 }
 
+/*
+ * call-seq:
+ *     Signal.signame(signo)  ->  string
+ *
+ *  convert signal number to signal name
+ *
+ *     Signal.trap("INT") { |signo| puts Signal.signame(signo) }
+ *     Process.kill("INT", 0)
+ *
+ *  <em>produces:</em>
+ *
+ *     INT
+ */
+static VALUE
+sig_signame(VALUE recv, VALUE signo)
+{
+    const char *signame = signo2signm(NUM2INT(signo));
+    return rb_str_new_cstr(signame);
+}
+
 const char *
 ruby_signal_name(int no)
 {
@@ -216,10 +241,11 @@ ruby_signal_name(int no)
 
 /*
  * call-seq:
- *    SignalException.new(sig)   =>  signal_exception
+ *    SignalException.new(sig_name)              ->  signal_exception
+ *    SignalException.new(sig_number [, name])   ->  signal_exception
  *
- *  Construct a new SignalException object.  +sig+ should be a known
- *  signal name, or a signal number.
+ *  Construct a new SignalException object.  +sig_name+ should be a known
+ *  signal name.
  */
 
 static VALUE
@@ -235,10 +261,7 @@ esignal_init(int argc, VALUE *argv, VALUE self)
 	if (!NIL_P(sig)) argnum = 2;
 	else sig = argv[0];
     }
-    if (argc < 1 || argnum < argc) {
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
-		 argc, argnum);
-    }
+    rb_check_arity(argc, 1, argnum);
     if (argnum == 2) {
 	signo = NUM2INT(sig);
 	if (signo < 0 || signo > NSIG) {
@@ -274,7 +297,7 @@ esignal_init(int argc, VALUE *argv, VALUE self)
 
 /*
  * call-seq:
- *    signal_exception.signo   =>  num
+ *    signal_exception.signo   ->  num
  *
  *  Returns a signal number.
  */
@@ -285,6 +308,7 @@ esignal_signo(VALUE self)
     return rb_iv_get(self, "signo");
 }
 
+/* :nodoc: */
 static VALUE
 interrupt_init(int argc, VALUE *argv, VALUE self)
 {
@@ -304,15 +328,15 @@ ruby_default_signal(int sig)
 
 /*
  *  call-seq:
- *     Process.kill(signal, pid, ...)    => fixnum
- *  
- *  Sends the given signal to the specified process id(s), or to the
- *  current process if _pid_ is zero. _signal_ may be an
- *  integer signal number or a POSIX signal name (either with or without
- *  a +SIG+ prefix). If _signal_ is negative (or starts
- *  with a minus sign), kills process groups instead of
+ *     Process.kill(signal, pid, ...)    -> fixnum
+ *
+ *  Sends the given signal to the specified process id(s) if _pid_ is positive.
+ *  If _pid_ is zero _signal_ is sent to all processes whose group ID is equal
+ *  to the group ID of the process. _signal_ may be an integer signal number or
+ *  a POSIX signal name (either with or without a +SIG+ prefix). If _signal_ is
+ *  negative (or starts with a minus sign), kills process groups instead of
  *  processes. Not all signals are available on all platforms.
- *     
+ *
  *     pid = fork do
  *        Signal.trap("HUP") { puts "Ouch!"; exit }
  *        # ... do some work ...
@@ -320,26 +344,37 @@ ruby_default_signal(int sig)
  *     # ...
  *     Process.kill("HUP", pid)
  *     Process.wait
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     Ouch!
+ *
+ *  If _signal_ is an integer but wrong for signal,
+ *  <code>Errno::EINVAL</code> or +RangeError+ will be raised.
+ *  Otherwise unless _signal_ is a +String+ or a +Symbol+, and a known
+ *  signal name, +ArgumentError+ will be raised.
+ *
+ *  Also, <code>Errno::ESRCH</code> or +RangeError+ for invalid _pid_,
+ *  <code>Errno::EPERM</code> when failed because of no privilege,
+ *  will be raised.  In these cases, signals may have been sent to
+ *  preceding processes.
  */
 
 VALUE
 rb_f_kill(int argc, VALUE *argv)
 {
 #ifndef HAS_KILLPG
-#define killpg(pg, sig) kill(-(pg), sig)
+#define killpg(pg, sig) kill(-(pg), (sig))
 #endif
     int negative = 0;
     int sig;
     int i;
+    volatile VALUE str;
     const char *s;
 
     rb_secure(2);
-    if (argc < 2)
-	rb_raise(rb_eArgError, "wrong number of arguments -- kill(sig, pid...)");
+    rb_check_arity(argc, 2, UNLIMITED_ARGUMENTS);
+
     switch (TYPE(argv[0])) {
       case T_FIXNUM:
 	sig = FIX2INT(argv[0]);
@@ -352,11 +387,11 @@ rb_f_kill(int argc, VALUE *argv)
 
       case T_STRING:
 	s = RSTRING_PTR(argv[0]);
+      str_signal:
 	if (s[0] == '-') {
 	    negative++;
 	    s++;
 	}
-      str_signal:
 	if (strncmp("SIG", s, 3) == 0)
 	    s += 3;
 	if((sig = signm2signo(s)) == 0)
@@ -367,17 +402,13 @@ rb_f_kill(int argc, VALUE *argv)
 	break;
 
       default:
-        {
-	    VALUE str;
-
-	    str = rb_check_string_type(argv[0]);
-	    if (!NIL_P(str)) {
-		s = RSTRING_PTR(str);
-		goto str_signal;
-	    }
-	    rb_raise(rb_eArgError, "bad signal type %s",
-		     rb_obj_classname(argv[0]));
+	str = rb_check_string_type(argv[0]);
+	if (!NIL_P(str)) {
+	    s = RSTRING_PTR(str);
+	    goto str_signal;
 	}
+	rb_raise(rb_eArgError, "bad signal type %s",
+		 rb_obj_classname(argv[0]));
 	break;
     }
 
@@ -408,7 +439,7 @@ static struct {
 #endif
 
 typedef RETSIGTYPE (*sighandler_t)(int);
-#ifdef SA_SIGINFO
+#ifdef USE_SIGALTSTACK
 typedef void ruby_sigaction_t(int, siginfo_t*, void*);
 #define SIGINFO_ARG , siginfo_t *info, void *ctx
 #else
@@ -416,38 +447,25 @@ typedef RETSIGTYPE ruby_sigaction_t(int);
 #define SIGINFO_ARG
 #endif
 
-#ifdef POSIX_SIGNAL
-#if defined(SIGSEGV) && defined(HAVE_SIGALTSTACK)
-#define USE_SIGALTSTACK
-#endif
-
 #ifdef USE_SIGALTSTACK
-#ifdef SIGSTKSZ
-#define ALT_STACK_SIZE (SIGSTKSZ*2)
-#else
-#define ALT_STACK_SIZE (4*1024)
-#endif
 /* alternate stack for SIGSEGV */
-static void
-register_sigaltstack(void)
+void
+rb_register_sigaltstack(rb_thread_t *th)
 {
-    static void *altstack = 0;
     stack_t newSS, oldSS;
 
-    if (altstack) return;
+    if (!th->altstack)
+	rb_bug("rb_register_sigaltstack: th->altstack not initialized\n");
 
-    newSS.ss_sp = altstack = malloc(ALT_STACK_SIZE);
-    if (newSS.ss_sp == NULL)
-	/* should handle error */
-	rb_bug("register_sigaltstack. malloc error\n");
+    newSS.ss_sp = th->altstack;
     newSS.ss_size = ALT_STACK_SIZE;
     newSS.ss_flags = 0;
 
-    if (sigaltstack(&newSS, &oldSS) < 0) 
-	rb_bug("register_sigaltstack. error\n");
+    sigaltstack(&newSS, &oldSS); /* ignore error. */
 }
-#endif
+#endif /* USE_SIGALTSTACK */
 
+#ifdef POSIX_SIGNAL
 static sighandler_t
 ruby_signal(int signum, sighandler_t handler)
 {
@@ -458,7 +476,7 @@ ruby_signal(int signum, sighandler_t handler)
 #endif
 
     sigemptyset(&sigact.sa_mask);
-#ifdef SA_SIGINFO
+#ifdef USE_SIGALTSTACK
     sigact.sa_sigaction = (ruby_sigaction_t*)handler;
     sigact.sa_flags = SA_SIGINFO;
 #else
@@ -471,11 +489,18 @@ ruby_signal(int signum, sighandler_t handler)
 	sigact.sa_flags |= SA_NOCLDWAIT;
 #endif
 #if defined(SA_ONSTACK) && defined(USE_SIGALTSTACK)
-    if (signum == SIGSEGV)
+    if (signum == SIGSEGV
+#ifdef SIGBUS
+	|| signum == SIGBUS
+#endif
+       )
 	sigact.sa_flags |= SA_ONSTACK;
 #endif
-    if (sigaction(signum, &sigact, &old) < 0)
-        rb_bug("sigaction error.\n");
+    if (sigaction(signum, &sigact, &old) < 0) {
+	if (errno != 0 && errno != EINVAL) {
+	    rb_bug_errno("sigaction", errno);
+	}
+    }
     return old.sa_handler;
 }
 
@@ -486,7 +511,7 @@ posix_signal(int signum, sighandler_t handler)
 }
 
 #else /* !POSIX_SIGNAL */
-#define ruby_signal(sig,handler) (/* rb_trap_accept_nativethreads[sig] = 0,*/ signal((sig),(handler)))
+#define ruby_signal(sig,handler) (/* rb_trap_accept_nativethreads[(sig)] = 0,*/ signal((sig),(handler)))
 #if 0 /* def HAVE_NATIVETHREAD */
 static sighandler_t
 ruby_nativethread_signal(int signum, sighandler_t handler)
@@ -505,39 +530,36 @@ sighandler(int sig)
 {
     ATOMIC_INC(signal_buff.cnt[sig]);
     ATOMIC_INC(signal_buff.size);
+    rb_thread_wakeup_timer_thread();
 #if !defined(BSD_SIGNAL) && !defined(POSIX_SIGNAL)
     ruby_signal(sig, sighandler);
 #endif
 }
 
-#if USE_TRAP_MASK
-# ifdef HAVE_SIGPROCMASK
-static sigset_t trap_last_mask;
-# else
-static int trap_last_mask;
-# endif
-#endif
+int
+rb_signal_buff_size(void)
+{
+    return signal_buff.size;
+}
 
 #if HAVE_PTHREAD_H
 #include <pthread.h>
 #endif
 
-void
+static void
 rb_disable_interrupt(void)
 {
-#ifndef _WIN32
+#ifdef HAVE_PTHREAD_SIGMASK
     sigset_t mask;
     sigfillset(&mask);
-    sigdelset(&mask, SIGVTALRM);
-    sigdelset(&mask, SIGSEGV);
     pthread_sigmask(SIG_SETMASK, &mask, NULL);
 #endif
 }
 
-void
+static void
 rb_enable_interrupt(void)
 {
-#ifndef _WIN32
+#ifdef HAVE_PTHREAD_SIGMASK
     sigset_t mask;
     sigemptyset(&mask);
     pthread_sigmask(SIG_SETMASK, &mask, NULL);
@@ -549,16 +571,14 @@ rb_get_next_signal(void)
 {
     int i, sig = 0;
 
-    for (i=1; i<RUBY_NSIG; i++) {
-	if (signal_buff.cnt[i] > 0) {
-	    rb_disable_interrupt();
-	    {
+    if (signal_buff.size != 0) {
+	for (i=1; i<RUBY_NSIG; i++) {
+	    if (signal_buff.cnt[i] > 0) {
 		ATOMIC_DEC(signal_buff.cnt[i]);
 		ATOMIC_DEC(signal_buff.size);
+		sig = i;
+		break;
 	    }
-	    rb_enable_interrupt();
-	    sig = i;
-	    break;
 	}
     }
     return sig;
@@ -566,8 +586,21 @@ rb_get_next_signal(void)
 
 #ifdef SIGBUS
 static RETSIGTYPE
-sigbus(int sig)
+sigbus(int sig SIGINFO_ARG)
 {
+/*
+ * Mac OS X makes KERN_PROTECTION_FAILURE when thread touch guard page.
+ * and it's delivered as SIGBUS instaed of SIGSEGV to userland. It's crazy
+ * wrong IMHO. but anyway we have to care it. Sigh.
+ */
+#if defined __APPLE__ && defined USE_SIGALTSTACK
+    int ruby_stack_overflowed_p(const rb_thread_t *, const void *);
+    NORETURN(void ruby_thread_stack_overflow(rb_thread_t *th));
+    rb_thread_t *th = GET_THREAD();
+    if (ruby_stack_overflowed_p(th, info->si_addr)) {
+	ruby_thread_stack_overflow(th);
+    }
+#endif
     rb_bug("Bus Error");
 }
 #endif
@@ -586,8 +619,8 @@ sigsegv(int sig SIGINFO_ARG)
     }
 #endif
     if (segv_received) {
-	fprintf(stderr, "SEGV recieved in SEGV handler\n");
-	exit(EXIT_FAILURE);
+	fprintf(stderr, "SEGV received in SEGV handler\n");
+	abort();
     }
     else {
 	extern int ruby_disable_gc_stress;
@@ -598,19 +631,27 @@ sigsegv(int sig SIGINFO_ARG)
 }
 #endif
 
-#ifdef SIGPIPE
-static RETSIGTYPE
-sigpipe(int sig)
-{
-    /* do nothing */
-}
-#endif
-
 static void
 signal_exec(VALUE cmd, int safe, int sig)
 {
-    VALUE signum = INT2NUM(sig);
-    rb_eval_cmd(cmd, rb_ary_new3(1, signum), safe);
+    rb_thread_t *cur_th = GET_THREAD();
+    volatile unsigned long old_interrupt_mask = cur_th->interrupt_mask;
+    int state;
+
+    cur_th->interrupt_mask |= TRAP_INTERRUPT_MASK;
+    TH_PUSH_TAG(cur_th);
+    if ((state = EXEC_TAG()) == 0) {
+	VALUE signum = INT2NUM(sig);
+	rb_eval_cmd(cmd, rb_ary_new3(1, signum), safe);
+    }
+    TH_POP_TAG();
+    cur_th = GET_THREAD();
+    cur_th->interrupt_mask = old_interrupt_mask;
+
+    if (state) {
+	/* XXX: should be replaced with rb_threadptr_async_errinfo_enque() */
+	JUMP_TAG(state);
+    }
 }
 
 void
@@ -655,31 +696,18 @@ rb_signal_exec(rb_thread_t *th, int sig)
 #ifdef SIGUSR2
 	  case SIGUSR2:
 #endif
-	    rb_thread_signal_raise(th, sig);
+	    rb_threadptr_signal_raise(th, sig);
 	    break;
 	}
     }
     else if (cmd == Qundef) {
-	rb_thread_signal_exit(th);
+	rb_threadptr_signal_exit(th);
     }
     else {
 	signal_exec(cmd, safe, sig);
     }
 }
 
-struct trap_arg {
-#if USE_TRAP_MASK
-# ifdef HAVE_SIGPROCMASK
-    sigset_t mask;
-# else
-    int mask;
-# endif
-#endif
-    int sig;
-    sighandler_t func;
-    VALUE cmd;
-};
-
 static sighandler_t
 default_handler(int sig)
 {
@@ -708,20 +736,20 @@ default_handler(int sig)
         break;
 #ifdef SIGBUS
       case SIGBUS:
-        func = sigbus;
+        func = (sighandler_t)sigbus;
         break;
 #endif
 #ifdef SIGSEGV
       case SIGSEGV:
         func = (sighandler_t)sigsegv;
 # ifdef USE_SIGALTSTACK
-        register_sigaltstack();
+        rb_register_sigaltstack(GET_THREAD());
 # endif
         break;
 #endif
 #ifdef SIGPIPE
       case SIGPIPE:
-        func = sigpipe;
+        func = SIG_IGN;
         break;
 #endif
       default:
@@ -743,6 +771,10 @@ trap_handler(VALUE *cmd, int sig)
     }
     else {
 	command = rb_check_string_type(*cmd);
+	if (NIL_P(command) && SYMBOL_P(*cmd)) {
+	    command = rb_id2str(SYM2ID(*cmd));
+	    if (!command) rb_raise(rb_eArgError, "bad handler");
+	}
 	if (!NIL_P(command)) {
 	    SafeStringValue(command);	/* taint check */
 	    *cmd = command;
@@ -786,6 +818,7 @@ sig_dfl:
 	else {
 	    rb_proc_t *proc;
 	    GetProcPtr(*cmd, proc);
+	    (void)proc;
 	}
     }
 
@@ -825,13 +858,17 @@ trap_signm(VALUE vsig)
 }
 
 static VALUE
-trap(struct trap_arg *arg)
+trap(int sig, sighandler_t func, VALUE command)
 {
-    sighandler_t oldfunc, func = arg->func;
-    VALUE oldcmd, command = arg->cmd;
-    int sig = arg->sig;
+    sighandler_t oldfunc;
+    VALUE oldcmd;
     rb_vm_t *vm = GET_VM();
 
+    /*
+     * Be careful. ruby_signal() and trap_list[sig].cmd must be changed
+     * atomically. In current implementation, we only need to don't call
+     * RUBY_VM_CHECK_INTS().
+     */
     oldfunc = ruby_signal(sig, func);
     oldcmd = vm->trap_list[sig].cmd;
     switch (oldcmd) {
@@ -847,48 +884,44 @@ trap(struct trap_arg *arg)
 
     vm->trap_list[sig].cmd = command;
     vm->trap_list[sig].safe = rb_safe_level();
-    /* enable at least specified signal. */
-#if USE_TRAP_MASK
-#ifdef HAVE_SIGPROCMASK
-    sigdelset(&arg->mask, sig);
-#else
-    arg->mask &= ~sigmask(sig);
-#endif
-#endif
+
     return oldcmd;
 }
 
-#if USE_TRAP_MASK
-static VALUE
-trap_ensure(struct trap_arg *arg)
+static int
+reserved_signal_p(int signo)
 {
-    /* enable interrupt */
-#ifdef HAVE_SIGPROCMASK
-    sigprocmask(SIG_SETMASK, &arg->mask, NULL);
-#else
-    sigsetmask(arg->mask);
+/* Synchronous signal can't deliver to main thread */
+#ifdef SIGSEGV
+    if (signo == SIGSEGV)
+	return 1;
 #endif
-    trap_last_mask = arg->mask;
-    return 0;
-}
+#ifdef SIGBUS
+    if (signo == SIGBUS)
+	return 1;
+#endif
+#ifdef SIGILL
+    if (signo == SIGILL)
+	return 1;
+#endif
+#ifdef SIGFPE
+    if (signo == SIGFPE)
+	return 1;
 #endif
 
-void
-rb_trap_restore_mask(void)
-{
-#if USE_TRAP_MASK
-# ifdef HAVE_SIGPROCMASK
-    sigprocmask(SIG_SETMASK, &trap_last_mask, NULL);
-# else
-    sigsetmask(trap_last_mask);
-# endif
+/* used ubf internal see thread_pthread.c. */
+#ifdef SIGVTALRM
+    if (signo == SIGVTALRM)
+	return 1;
 #endif
+
+    return 0;
 }
 
 /*
  * call-seq:
- *   Signal.trap( signal, command ) => obj
- *   Signal.trap( signal ) {| | block } => obj
+ *   Signal.trap( signal, command ) -> obj
+ *   Signal.trap( signal ) {| | block } -> obj
  *
  * Specifies the handling of signals. The first parameter is a signal
  * name (a string such as ``SIGALRM'', ``SIGUSR1'', and so on) or a
@@ -919,44 +952,41 @@ rb_trap_restore_mask(void)
 static VALUE
 sig_trap(int argc, VALUE *argv)
 {
-    struct trap_arg arg;
+    int sig;
+    sighandler_t func;
+    VALUE cmd;
 
     rb_secure(2);
-    if (argc == 0 || argc > 2) {
-	rb_raise(rb_eArgError, "wrong number of arguments -- trap(sig, cmd)/trap(sig){...}");
+    rb_check_arity(argc, 1, 2);
+
+    sig = trap_signm(argv[0]);
+    if (reserved_signal_p(sig)) {
+        const char *name = signo2signm(sig);
+        if (name)
+            rb_raise(rb_eArgError, "can't trap reserved signal: SIG%s", name);
+        else
+            rb_raise(rb_eArgError, "can't trap reserved signal: %d", sig);
     }
 
-    arg.sig = trap_signm(argv[0]);
     if (argc == 1) {
-	arg.cmd = rb_block_proc();
-	arg.func = sighandler;
+	cmd = rb_block_proc();
+	func = sighandler;
     }
-    else if (argc == 2) {
-	arg.cmd = argv[1];
-	arg.func = trap_handler(&arg.cmd, arg.sig);
+    else {
+	cmd = argv[1];
+	func = trap_handler(&cmd, sig);
     }
 
-    if (OBJ_TAINTED(arg.cmd)) {
+    if (OBJ_TAINTED(cmd)) {
 	rb_raise(rb_eSecurityError, "Insecure: tainted signal trap");
     }
-#if USE_TRAP_MASK
-    /* disable interrupt */
-# ifdef HAVE_SIGPROCMASK
-    sigfillset(&arg.mask);
-    sigprocmask(SIG_BLOCK, &arg.mask, &arg.mask);
-# else
-    arg.mask = sigblock(~0);
-# endif
 
-    return rb_ensure(trap, (VALUE)&arg, trap_ensure, (VALUE)&arg);
-#else
-    return trap(&arg);
-#endif
+    return trap(sig, func, cmd);
 }
 
 /*
  * call-seq:
- *   Signal.list => a_hash
+ *   Signal.list -> a_hash
  *
  * Returns a list of signal names mapped to the corresponding
  * underlying signal numbers.
@@ -980,10 +1010,14 @@ install_sighandler(int signum, sighandler_t handler)
 {
     sighandler_t old;
 
+    /* At this time, there is no subthread. Then sigmask guarantee atomics. */
+    rb_disable_interrupt();
     old = ruby_signal(signum, handler);
+    /* signal handler should be inherited during exec. */
     if (old != SIG_DFL) {
 	ruby_signal(signum, old);
     }
+    rb_enable_interrupt();
 }
 
 #if defined(SIGCLD) || defined(SIGCHLD)
@@ -991,46 +1025,20 @@ static void
 init_sigchld(int sig)
 {
     sighandler_t oldfunc;
-#if USE_TRAP_MASK
-# ifdef HAVE_SIGPROCMASK
-    sigset_t mask;
-# else
-    int mask;
-# endif
-#endif
-
-#if USE_TRAP_MASK
-    /* disable interrupt */
-# ifdef HAVE_SIGPROCMASK
-    sigfillset(&mask);
-    sigprocmask(SIG_BLOCK, &mask, &mask);
-# else
-    mask = sigblock(~0);
-# endif
-#endif
 
+    rb_disable_interrupt();
     oldfunc = ruby_signal(sig, SIG_DFL);
     if (oldfunc != SIG_DFL && oldfunc != SIG_IGN) {
 	ruby_signal(sig, oldfunc);
     } else {
 	GET_VM()->trap_list[sig].cmd = 0;
     }
-
-#if USE_TRAP_MASK
-#ifdef HAVE_SIGPROCMASK
-    sigdelset(&mask, sig);
-    sigprocmask(SIG_SETMASK, &mask, NULL);
-#else
-    mask &= ~sigmask(sig);
-    sigsetmask(mask);
-#endif
-    trap_last_mask = mask;
-#endif
+    rb_enable_interrupt();
 }
 #endif
 
 void
-ruby_sig_finalize()
+ruby_sig_finalize(void)
 {
     sighandler_t oldfunc;
 
@@ -1041,8 +1049,9 @@ ruby_sig_finalize()
 }
 
 
-#ifdef RUBY_DEBUG_ENV
 int ruby_enable_coredump = 0;
+#ifndef RUBY_DEBUG_ENV
+#define ruby_enable_coredump 0
 #endif
 
 /*
@@ -1090,6 +1099,7 @@ Init_signal(void)
     rb_define_global_function("trap", sig_trap, -1);
     rb_define_module_function(mSignal, "trap", sig_trap, -1);
     rb_define_module_function(mSignal, "list", sig_list, 0);
+    rb_define_module_function(mSignal, "signame", sig_signame, 1);
 
     rb_define_method(rb_eSignal, "initialize", esignal_init, -1);
     rb_define_method(rb_eSignal, "signo", esignal_signo, 0);
@@ -1116,22 +1126,19 @@ Init_signal(void)
     install_sighandler(SIGUSR2, sighandler);
 #endif
 
-#ifdef RUBY_DEBUG_ENV
-    if (!ruby_enable_coredump)
-#endif
-    {
+    if (!ruby_enable_coredump) {
 #ifdef SIGBUS
-    install_sighandler(SIGBUS, sigbus);
+	install_sighandler(SIGBUS, (sighandler_t)sigbus);
 #endif
 #ifdef SIGSEGV
 # ifdef USE_SIGALTSTACK
-    register_sigaltstack();
+	rb_register_sigaltstack(GET_THREAD());
 # endif
-    install_sighandler(SIGSEGV, (sighandler_t)sigsegv);
+	install_sighandler(SIGSEGV, (sighandler_t)sigsegv);
 #endif
     }
 #ifdef SIGPIPE
-    install_sighandler(SIGPIPE, sigpipe);
+    install_sighandler(SIGPIPE, SIG_IGN);
 #endif
 
 #if defined(SIGCLD)
